@@ -1,3 +1,4 @@
+"""Scala interpreter supporting async I/O with the managing Python process."""
 import asyncio
 import atexit
 import logging
@@ -8,43 +9,49 @@ import signal
 import tempfile
 
 from asyncio import Future
+from collections import namedtuple
 from concurrent.futures import ThreadPoolExecutor
 from typing import Callable, Union, List, Any
 
 import spylon.spark
 
-spark_session = None
-spark_jvm_helpers = None
+# Global singletons
+SparkState = namedtuple('SparkState', 'spark_session spark_jvm_helpers')
+spark_state = None
 scala_intp = None
 
+# Default Spark application name
+DEFAULT_APPLICATION_NAME = "spylon-kernel"
 
-def init_spark_session(conf: spylon.spark.SparkConfiguration=None,
-                       application_name: str="ScalaMetaKernel"):
-    """Initialize the Spark session.
+
+def init_spark(conf=None, application_name=None):
+    """Initializes a SparkSession
 
     Parameters
     ----------
-    conf: optional
+    conf: spylon.spark.SparkConfiguration, optional
         Spark configuration to apply to the session
-    application_name: optional
+    application_name: str, optional
         Name to give the session
 
-
+    Returns
+    -------
+    SparkState namedtuple
     """
+    global spark_state
+    # If we have already initialized a spark session stop
+    if spark_state:
+        return spark_state
+
     # Ensure we have the correct classpath settings for the repl to work.
     os.environ.setdefault('SPARK_SUBMIT_OPTS', '-Dscala.usejavacp=true')
 
-    global spark_session
-    # If we have already initialized a spark session. Don't carry on.
-    if spark_session:
-        return
     if conf is None:
         conf = spylon.spark.launcher.SparkConfiguration()
-    # SparkContext will detect this configuration and register it with the RpcEnv's
-    # file server, setting spark.repl.class.uri to the actual URI for executors to
-    # use. This is sort of ugly but since executors are started as part of SparkContext
-    # initialization in certain cases, there's an initialization order issue that prevents
-    # this from being set after SparkContext is instantiated.
+    if application_name is None:
+        application_name = DEFAULT_APPLICATION_NAME
+
+    # Create a temp directory that gets cleaned up on exit
     output_dir = os.path.abspath(tempfile.mkdtemp())
 
     def cleanup():
@@ -52,19 +59,29 @@ def init_spark_session(conf: spylon.spark.SparkConfiguration=None,
 
     atexit.register(cleanup)
     signal.signal(signal.SIGTERM, cleanup)
+
+    # SparkContext will detect this configuration and register it with the RpcEnv's
+    # file server, setting spark.repl.class.uri to the actual URI for executors to
+    # use. This is sort of ugly but since executors are started as part of SparkContext
+    # initialization in certain cases, there's an initialization order issue that prevents
+    # this from being set after SparkContext is instantiated.
     conf.conf.set("spark.repl.class.outputDir", output_dir)
 
+    # Create a new spark context using the configuration
     spark_context = conf.spark_context(application_name)
-    from pyspark.sql import SparkSession
-    spark_session = SparkSession(spark_context)
-    from spylon.spark.utils import SparkJVMHelpers
-    global spark_jvm_helpers
-    # noinspection PyProtectedMember
-    spark_jvm_helpers = SparkJVMHelpers(spark_session._sc)
 
+    # pyspark is in the python path after create the context
+    from pyspark.sql import SparkSession
+    from spylon.spark.utils import SparkJVMHelpers
+
+    # Create the singleton SparkState
+    spark_session = SparkSession(spark_context)
+    spark_jvm_helpers = SparkJVMHelpers(spark_session._sc)
+    spark_state = SparkState(spark_session, spark_jvm_helpers)
+    return spark_state
 
 def get_web_ui_url(sc):
-    """Get the web ui for a spark context
+    """Gets the URL of the Spark web UI for a given SparkContext.
 
     Parameters
     ----------
@@ -73,6 +90,7 @@ def get_web_ui_url(sc):
     Returns
     -------
     url : str
+        URL to the Spark web UI
     """
     # Dig into the java spark conf to actually be able to resolve the spark configuration
     # noinspection PyProtectedMember
@@ -107,80 +125,93 @@ def initialize_scala_interpreter():
 
     Notes
     -----
-    Portions of this have been adapted out of Apache Toree and Apache Zeppelin
+    Portions of this have been adapted out of Apache Toree and Apache Zeppelin.
 
     Returns
     -------
-    SparkInterpreter
+    ScalaInterpreter
     """
-    if spark_session is None:
-        init_spark_session()
-
-    from spylon.spark.utils import SparkJVMHelpers
-    assert isinstance(spark_jvm_helpers, SparkJVMHelpers)
-    from pyspark.sql import SparkSession
-    assert isinstance(spark_session, SparkSession)
+    # Initialize Spark first if it isn't already
+    spark_session, spark_jvm_helpers = init_spark()
 
+    # Get handy JVM references
     jvm = spark_session._jvm
     jconf = spark_session._jsc.getConf()
-    bytes_out = jvm.org.apache.commons.io.output.ByteArrayOutputStream()
-
     io = jvm.java.io
 
-    jprintWriter = io.PrintWriter(bytes_out, True)
+    # Build a print writer that'll be used to get results from the
+    # Scala REPL
+    bytes_out = jvm.org.apache.commons.io.output.ByteArrayOutputStream()
+    jprint_writer = io.PrintWriter(bytes_out, True)
 
-    execUri = jvm.System.getenv("SPARK_EXECUTOR_URI")
-    jconf.setIfMissing("spark.app.name", "Spark shell")
-    if (execUri is not None):
-        jconf.set("spark.executor.uri", execUri)
+    # Set the Spark applicaiton name if it is not already set
+    jconf.setIfMissing("spark.app.name", DEFAULT_APPLICATION_NAME)
+
+    # Set the location of the Spark package on HDFS, if available
+    exec_uri = jvm.System.getenv("SPARK_EXECUTOR_URI")
+    if exec_uri is not None:
+        jconf.set("spark.executor.uri", exec_uri)
+
+    # Configure the classpath and temp directory created by init_spark
+    # as the shared path for REPL generated artifacts
     output_dir = jconf.get("spark.repl.class.outputDir")
-
     jars = jvm.org.apache.spark.util.Utils.getUserJars(jconf, True).mkString(":")
-    interpArguments = spark_jvm_helpers.to_scala_list(
+    interp_arguments = spark_jvm_helpers.to_scala_list(
         ["-Yrepl-class-based", "-Yrepl-outdir", output_dir,
-         "-classpath", jars
+         "-classpath", jars, "-deprecation:false"
          ]
     )
-
     settings = jvm.scala.tools.nsc.Settings()
-    settings.processArguments(interpArguments, True)
+    settings.processArguments(interp_arguments, True)
 
-    # Since we have already instantiated our spark context on the python side,
-    # set it in the Main repl class as well
+    # Since we have already instantiated our SparkSession on the Python side,
+    # share it with the Scala Main REPL class as well
     Main = jvm.org.apache.spark.repl.Main
     jspark_session = spark_session._jsparkSession
-    # equivalent to Main.sparkSession = jspark_session
+    # Equivalent to Main.sparkSession = jspark_session
     getattr(Main, "sparkSession_$eq")(jspark_session)
     getattr(Main, "sparkContext_$eq")(jspark_session.sparkContext())
 
-    def start_imain():
-        intp = jvm.scala.tools.nsc.interpreter.IMain(settings, jprintWriter)
-        intp.initializeSynchronous()
+    # Instantiate a Scala interpreter
+    intp = jvm.scala.tools.nsc.interpreter.IMain(settings, jprint_writer)
+    intp.initializeSynchronous()
 
-        # Ensure that sc and spark are bound in the interpreter context.
-        intp.interpret("""
-            @transient val spark = if (org.apache.spark.repl.Main.sparkSession != null) {
-                org.apache.spark.repl.Main.sparkSession
-              } else {
-                org.apache.spark.repl.Main.createSparkSession()
-              }
-            @transient val sc = {
-              val _sc = spark.sparkContext
-              _sc
-            }
-            """)
-        intp.interpret("import org.apache.spark.SparkContext._")
-        intp.interpret("import spark.implicits._")
-        intp.interpret("import spark.sql")
-        intp.interpret("import org.apache.spark.sql.functions._")
-        bytes_out.reset()
-        return intp
+    # Ensure that sc and spark are bound in the interpreter context.
+    intp.interpret("""
+        @transient val spark = if (org.apache.spark.repl.Main.sparkSession != null) {
+            org.apache.spark.repl.Main.sparkSession
+        } else {
+            org.apache.spark.repl.Main.createSparkSession()
+        }
+        @transient val sc = {
+            val _sc = spark.sparkContext
+            _sc
+        }
+        """)
+    # Import Spark packages for convenience
+    intp.interpret("import org.apache.spark.SparkContext._")
+    intp.interpret("import spark.implicits._")
+    intp.interpret("import spark.sql")
+    intp.interpret("import org.apache.spark.sql.functions._")
+    # Clear the print writer stream
+    bytes_out.reset()
 
-    imain = start_imain()
-    return SparkInterpreter(jvm, imain, bytes_out)
+    return ScalaInterpreter(jvm, intp, bytes_out)
 
 
 def _scala_seq_to_py(jseq):
+    """Generator for all elements in a Scala sequence.
+
+    Parameters
+    ----------
+    jseq : Scala Seq
+        Scala sequence
+
+    Yields
+    ------
+    any
+        One element per sequence
+    """
     n = jseq.size()
     for i in range(n):
         yield jseq.apply(i)
@@ -192,11 +223,8 @@ class ScalaException(Exception):
         self.scala_message = scala_message
 
 
-tOutputHandler = Callable[[List[Any]], None]
-
-
-class SparkInterpreter(object):
-    """Wrapper for a scala interpreter.
+class ScalaInterpreter(object):
+    """Wrapper for a Scala interpreter.
 
     Notes
     -----
@@ -210,107 +238,159 @@ class SparkInterpreter(object):
     jbyteout : py4j.java_gateway.JavaObject
         Java object representing an instance of `org.apache.commons.io.output.ByteArrayOutputStream`
         This is used to return output data from the REPL.
+    log : logging.Logger
+        Logger for this instance
     loop : asyncio.AbstractEventLoop, optional
         Asyncio eventloop
-
+    web_ui_url : str
+        URL of the Spark web UI associated with this interpreter
     """
-    executor = ThreadPoolExecutor(4)
+    executor = ThreadPoolExecutor(1)
 
     def __init__(self, jvm, jimain, jbyteout, loop: Union[None, asyncio.AbstractEventLoop]=None):
-        self.spark_session = spark_session
-        # noinspection PyProtectedMember
-        self.sc = spark_session._sc
-        self.web_ui_url = get_web_ui_url(self.sc)
-        self._jcompleter = None
         self.jvm = jvm
         self.jimain = jimain
-        if loop is None:
-            # TODO: We may want to use new_event_loop here to avoid stopping and starting the main one.
-            loop = asyncio.get_event_loop()
-        self.loop = loop
+        self.jbyteout = jbyteout
         self.log = logging.getLogger(self.__class__.__name__)
 
-        jinterpreter_package = getattr(getattr(self.jvm.scala.tools.nsc.interpreter, 'package$'), "MODULE$")
-        self.iMainOps = jinterpreter_package.IMainOps(jimain)
-        self.jbyteout = jbyteout
+        if loop is None:
+            # TODO: We may want to use new_event_loop here to avoid stopping
+            # and starting the main one.
+            loop = asyncio.get_event_loop()
+        self.loop = loop
 
+        # Store the state here so that clients of the instance
+        # can access them (for now ...)
+        self.sc = spark_state.spark_session._sc
+        self.spark_session = spark_state.spark_session
+
+        # noinspection PyProtectedMember
+        self.web_ui_url = get_web_ui_url(self.sc)
+        self._jcompleter = None
+
+        # ???
+        # jinterpreter_package = getattr(getattr(self.jvm.scala.tools.nsc.interpreter, 'package$'), "MODULE$")
+        # self.iMainOps = jinterpreter_package.IMainOps(jimain)
+
+        # Create a temp directory that will contain the stdout/stderr
+        # files written by Scala and read by Python
         tempdir = tempfile.mkdtemp()
         atexit.register(shutil.rmtree, tempdir, True)
-        self.tempdir = tempdir
-        # Handlers for dealing with stout and stderr.  This allows us to insert additional behavior for magics
+
+        # Handlers for dealing with stdout and stderr.
         self._stdout_handlers = []
         self._stderr_handlers = []
-        self._initialize_stdout_err()
+        self._initialize_stdout_err(tempdir)
 
-    def register_stdout_handler(self, handler: tOutputHandler):
+    def register_stdout_handler(self, handler):
+        """Registers a handler for the Scala stdout stream.
+
+        Parameters
+        ----------
+        handler: callable(str) -> None
+            Function to handle a stdout from the interpretter
+        """
         self._stdout_handlers.append(handler)
 
-    def register_stderr_handler(self, handler: tOutputHandler):
+    def register_stderr_handler(self, handler):
+        """Registers a handler for the Scala stderr stream.
+
+        Parameters
+        ----------
+        handler: callable(str) -> None
+            Function to handle a stdout from the interpretter
+        """
         self._stderr_handlers.append(handler)
 
-    def _initialize_stdout_err(self):
-        stdout_file = os.path.abspath(os.path.join(self.tempdir, 'stdout'))
-        stderr_file = os.path.abspath(os.path.join(self.tempdir, 'stderr'))
-        # Start up the pipes on the JVM side
+    def _initialize_stdout_err(self, tempdir):
+        """Redirects stdout/stderr in the Scala interpreter to two files in the
+        given tempdir and begins async tasks to poll those files for lines of text.
 
-        self.log.info("Before Java redirected")
-        self.log.debug("stdout/err redirected to %s", self.tempdir)
-        code = 'Console.set{pipe}(new PrintStream(new FileOutputStream(new File(new java.net.URI("{filename}")), true)))'
+        Parameters
+        ----------
+        tempdir : str
+            Temporary directory
+        """
+        stdout_file = os.path.abspath(os.path.join(tempdir, 'stdout'))
+        stderr_file = os.path.abspath(os.path.join(tempdir, 'stderr'))
+
+        code = 'val _std{pipe} = new PrintStream(new FileOutputStream(new File(new java.net.URI("{filename}")), true))'
         code = '\n'.join([
             'import java.io.{PrintStream, FileOutputStream, File}',
             'import scala.Console',
-            code.format(pipe="Out", filename=pathlib.Path(stdout_file).as_uri()),
-            code.format(pipe="Err", filename=pathlib.Path(stderr_file).as_uri())
+            code.format(pipe="out", filename=pathlib.Path(stdout_file).as_uri()),
+            code.format(pipe="err", filename=pathlib.Path(stderr_file).as_uri())
         ])
-        o = self.interpret(code)
-        self.log.info("Console redirected, %s", o)
+        self.interpret(code, redirect_streams=False)
 
         self.loop.create_task(self._poll_file(stdout_file, self.handle_stdout))
         self.loop.create_task(self._poll_file(stderr_file, self.handle_stderr))
 
-    def handle_stdout(self, *args) -> None:
-        for handler in self._stdout_handlers:
-            handler(*args)
+    def handle_stdout(self, line):
+        """Passes a line of Scala stdout to registered handlers.
 
-    def handle_stderr(self, *args) -> None:
-        for handler in self._stderr_handlers:
-            handler(*args)
-
-    async def _poll_file(self, filename: str, fn: Callable[[Any], None]):
+        Parameters
+        ----------
+        line : str
+            Line of text
         """
+        for handler in self._stdout_handlers:
+            handler(line)
+
+    def handle_stderr(self, line):
+        """Passes a line of Scala stderr to registered handlers.
+
+        Parameters
+        ----------
+        line : str
+            Line of text
+        """
+        for handler in self._stderr_handlers:
+            handler(line)
+
+    async def _poll_file(self, filename, fn):
+        """Busy-polls a file for lines of text and passes them to
+        the provided callback function when available.
 
         Parameters
         ----------
         filename : str
-        fn : (str) -> None
-            Function to deal with string output.
+            Filename to poll for lines of text
+        fn : callable(str) -> None
+            Callback function that handles lines of text
         """
         fd = open(filename, 'r')
         while True:
-            line = fd.readline()
-            if line:
-                # processing a line from the file and running our processing function.
-                fn(line)
-                # self.log.critical("AFTER PUSH")
+            chars = fd.read(4096)
+            if chars:
+                fn(chars)
                 await asyncio.sleep(0, loop=self.loop)
             else:
                 await asyncio.sleep(0.01, loop=self.loop)
 
-    def _interpret_sync(self, code: str, synthetic=False):
-        """Interpret a block of scala code.
+    def _interpret_sync(self, code, synthetic=False):
+        """Synchronously interprets a Block of scala code and returns the
+        string output from the Scala REPL.
 
-        If you want to get the result as a python object, follow this with a
-        call to `last_result()`
+        If you want to get the result as a Python object, follow this with a
+        call to `last_result`.
 
         Parameters
         ----------
         code : str
-        synthetic : bool
+            Scala code to interpret
+        synthetic : bool, optional
+            Use synthetic Scala classes (?)
 
         Returns
         -------
-        reploutput : str
-            String output from the scala REPL.
+        str
+            String output from the scala REPL
+
+        Raises
+        ------
+        ScalaException
+            When there is a problem interpreting the code
         """
         try:
             res = self.jimain.interpret(code, synthetic)
@@ -325,51 +405,66 @@ class SparkInterpreter(object):
             elif result == 'Error':
                 raise ScalaException(pyres)
             elif result == 'Incomplete':
-                raise ScalaException(pyres)
+                raise ScalaException(pyres or '<console>: error: incomplete input')
             return pyres
         finally:
             self.jbyteout.reset()
 
-    async def _interpret_async(self, code: str, future: Future):
-        """Async execute for running a block of scala code.
+    async def _interpret_async(self, code, future):
+        """Asynchronously interprets a block of Scala code and sets the
+        output or exception as the result of the future.
 
         Parameters
         ----------
         code : str
+            Scala code to interpret
         future : Future
-            future used to hold the result of the computation.
+            Future result or exception
         """
         try:
             result = await self.loop.run_in_executor(self.executor, self._interpret_sync, code)
             future.set_result(result)
         except Exception as e:
             future.set_exception(e)
-        return
 
-    def interpret(self, code: str):
-        """Interpret a block of scala code.
+    def interpret(self, code, redirect_streams=True):
+        """Interprets a block of Scala code.
 
-        If you want to get the result as a python object, follow this will a call to `last_result()`
+        Follow this with a call `last_result` to retrieve the result as a
+        Python object.
 
         Parameters
         ----------
         code : str
+            Scala code to interpret
+        redirect_streams : bool, optional
+            Force stdout/stderr to the polled files before executing
+            `code`
 
         Returns
         -------
-        reploutput : str
-            String output from the scala REPL.
+        str
+            String output from the scala REPL
+
+        Raises
+        ------
+        ScalaException
+            When there is a problem interpreting the code
         """
+        if redirect_streams:
+            # Ensure stdout and stderr are going to the temp files before
+            # executing the code.
+            code = 'Console.setOut(_stdout)\nConsole.setErr(_stderr)\n' + code
         fut = asyncio.Future(loop=self.loop)
         asyncio.ensure_future(self._interpret_async(code, fut), loop=self.loop)
         res = self.loop.run_until_complete(fut)
         return res
 
     def last_result(self):
-        """Retrieves the jvm result object from the previous call to interpret.
+        """Retrieves the JVM result object from the preceeding call to `interpret`.
 
-        If the result is a supported primitive type it is converted to a python object, otherwise it returns a py4j
-        view onto that object.
+        If the result is a supported primitive type, convers it to a Python object.
+        Otherwise, returns a py4j view onto that object.
 
         Returns
         -------
@@ -377,11 +472,11 @@ class SparkInterpreter(object):
         """
         # TODO : when evaluating multiline expressions this returns the first result
         lr = self.jimain.lastRequest()
-        res = lr.lineRep().call("$result", spark_jvm_helpers.to_scala_list([]))
+        res = lr.lineRep().call("$result", spark_state.spark_jvm_helpers.to_scala_list([]))
         return res
 
-    def bind(self, name: str, value: Any, jtyp: str="Any"):
-        """Set a variable in the scala repl environment to a python valued type.
+    def bind(self, name, value, jtyp="Any"):
+        """Set a variable in the Scala REPL to a Python valued type.
 
         Parameters
         ----------
@@ -390,8 +485,12 @@ class SparkInterpreter(object):
         jtyp : str
             String representation of the Java type that we want to cast this as.
 
+        Returns
+        -------
+        bool
+            True if the value is of one of the compatible types, False if not
         """
-        modifiers = spark_jvm_helpers.to_scala_list(["@transient"])
+        modifiers = spark_state.spark_jvm_helpers.to_scala_list(["@transient"])
         # Ensure that the value that we are trying to set here is a compatible type on the java side
         # Import is here due to lazily instantiating the SparkContext
         from py4j.java_gateway import JavaClass, JavaObject, JavaMember
@@ -399,17 +498,25 @@ class SparkInterpreter(object):
             int, str, bytes, bool, list, dict, JavaClass, JavaMember, JavaObject
         )
         if isinstance(value, compatible_types):
-            self.jimain.bind(name, "Any", value, modifiers)
+            self.jimain.bind(name, jtyp, value, modifiers)
+            return True
+        return False
 
     @property
     def jcompleter(self):
+        """Scala code completer.
+
+        Returns
+        -------
+        scala.tools.nsc.interpreter.PresentationCompilerCompleter
+        """
         if self._jcompleter is None:
             jClass = self.jvm.scala.tools.nsc.interpreter.PresentationCompilerCompleter
             self._jcompleter = jClass(self.jimain)
         return self._jcompleter
 
-    def complete(self, code: str, pos: int) -> List[str]:
-        """Performs code completion for a block of scala code.
+    def complete(self, code, pos):
+        """Performs code completion for a block of Scala code.
 
         Parameters
         ----------
@@ -421,17 +528,19 @@ class SparkInterpreter(object):
         Returns
         -------
         List[str]
+            Candidates for code completion
         """
         c = self.jcompleter
         jres = c.complete(code, pos)
         return list(_scala_seq_to_py(jres.candidates()))
 
     def is_complete(self, code):
-        """Determine if a hunk of code is a complete block of scala.
+        """Determines if a chunk of code is a complete block of Scala.
 
         Parameters
         ----------
         code : str
+            Code to evaluate for completeness
 
         Returns
         -------
@@ -448,26 +557,26 @@ class SparkInterpreter(object):
                 return 'incomplete'
             else:
                 return 'invalid'
-
         finally:
             self.jbyteout.reset()
 
-    def get_help_on(self, info):
-        """For a given symbol attempt to get some minor help on it in terms of function signature.
+    def get_help_on(self, obj):
+        """Gets the signature for the given object.
 
-        Due to the JVM having no runtime docstring information, the level of detail we can retrieve is rather limited.
+        Due to the JVM having no runtime docstring information, the level of
+        detail is rather limited.
 
         Parameters
         ----------
-        info : str
-            object name to try and get information for
+        obj : str
+            Object to fetch info about
 
         Returns
         -------
         str
-
+            typeAt hint from Scala
         """
-        code = info + '// typeAt {} {}'.format(0, len(info))
+        code = obj + '// typeAt {} {}'.format(0, len(obj))
         scala_type = self.complete(code, len(code))
         # When using the // typeAt hint we will get back a list made by
         # "" :: type :: Nil
@@ -477,10 +586,6 @@ class SparkInterpreter(object):
         #       better results for the function in question
         return scala_type[-1]
 
-    def printHelp(self):
-        return self.jimain.helpSummary()
-
-
 def get_scala_interpreter():
     """Get the scala interpreter instance.
 
@@ -488,7 +593,7 @@ def get_scala_interpreter():
 
     Returns
     -------
-    scala_intp : SparkInterpreter
+    scala_intp : ScalaInterpreter
     """
     global scala_intp
     if scala_intp is None:
